/*
 * Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 * AMAZON.COM CONFIDENTIAL
 */

/**
 * This needs to be a singleton across all webviews to avoid things getting out of sync.
 * Handles the FTUE workflow and related settings.
 * Authoritative source for determining whether or not user data can be uploaded.
 * TODO: Add retries on network calls.
 * TODO: Handle broken token issues?
 */
FTUEHandler = (function() {
    var constructor = function(profilesHandler, countries, gcSettings) {
        var thisHandler = this;
        
        /* FTUE setting for whether or not the FTUE has been completed and the game is connected. */
        var SETTINGS_FTUE_CONNECTED = "ftue:connected";
        
        /* User setting for whether or not they want to be in guest mode. */
        var SETTINGS_USER_GUEST = "user:guest";
        
        /* User's country of residence. */
        var SETTINGS_USER_COR = "user:cor";

        var requestTypeHandled = 'FTUERequest';

        /** If true, upsell is allowed */
        var upsellEnabled = false;
        
        /** If true, on startup the FTUE workflow will check if connection is possible without prompting the user. */
        var autoConnectEnabled = false;
        
        /** If true, the user's cached cor is blank. */
        var cachedCORIsBlank = false;

        /** If true, the user's cached cor is not supported. */
        var cachedCORIsNotSupported = false;
        
        /**
         * Resolved when everything is initialized and internal state is generally stable,
         * at least until the user does something.
         */
        var initializationDeferred = $.Deferred();
        
        /* ========================================== Initialization ========================================== */
        
        /**
         * Reloads the internal state based on stuff that can be obtained without making a network call.
         *
         * States that will be persisted to disk via the settings above:
         *                       can't upload data (guest) | upsell | auto connect | notes
         *     COR not supported    Y                         N          N          Check this first, and verify still logged in
         *     user guest mode      Y                         Y          N          Check this second
         *     ftue connected       N                         N         n/a         Check this third, and verify still logged in
         *     basic guest mode     Y                         Y          Y          Otherwise
         */
        var reloadInternalState = function() {
            var dfd = $.Deferred();
            gcSettings.getUserCOR().always(function(corSupported) {
                cachedCORIsBlank = corSupported == null || corSupported == '';
                cachedCORIsNotSupported = !cachedCORIsBlank && !corSupported;
                
                if (cachedCORIsNotSupported) {
                    getLoggedInStatus().always(function(isLoggedIn) {
                        if (isLoggedIn) {
                            console.log("FTUE reload state: COR not supported");
                            upsellEnabled = false;
                            autoConnectEnabled = false;
                            gcSettings.setGuestMode(true).always(function() {
                                dfd.resolve();
                            });
                        } else {
                            console.log("FTUE reload state: Unexpected logout, returning to guest mode.");
                            gcSettings.setFTUEConnected(false).always(function() {
                                gcSettings.setUserCOR(null).always(function() {
                                    upsellEnabled = true;
                                    autoConnectEnabled = true;
                                    gcSettings.setGuestMode(true).always(function() {
                                        dfd.resolve();
                                    });
                                });
                            });
                        }
                    });
                } else {
                    gcSettings.getFTUEConnected().always(function(cachedFTUEConnected) {
                        if (cachedFTUEConnected) {
                            getLoggedInStatus().always(function(isLoggedIn) {
                                if (isLoggedIn) {
                                    console.log("FTUE reload state: connected");
                                    upsellEnabled = false;
                                    autoConnectEnabled = false;
                                    gcSettings.setGuestMode(false).always(function() {
                                        dfd.resolve();
                                    });
                                } else {
                                    /**
                                     * They unexpectedly logged out.
                                     * This could happen if the game integrates with IBA and logs them out.
                                     * This is also triggered by revoking using the IBA page on the retail website
                                     *     under My Account, Settings, Login with Amazon connections.
                                     */
                                    /* TODO: Are we going to have a logout button?  If so, ftue needs to be involved/updated. */
                                    /* TODO: Detect revoke and switch to guest mode during a session.  Maybe on overlay launch? */
                                    console.log("FTUE reload state: Unexpected logout, returning to guest mode.");
                                    gcSettings.setFTUEConnected(false).always(function() {
                                        gcSettings.setUserCOR(null).always(function() {
                                            upsellEnabled = true;
                                            autoConnectEnabled = true;
                                            gcSettings.setGuestMode(true).always(function() {
                                                dfd.resolve();
                                            });
                                        });
                                    });
                                }
                            });
                        } else {
                            console.log("FTUE reload state: guest");
                            upsellEnabled = true;
                            autoConnectEnabled = true;
                            gcSettings.setGuestMode(true).always(function() {
                                dfd.resolve();
                            });
                        }
                    });
                }
            });
            return dfd.promise();
        };
        
        /*
        TODO: Hide welcome back toast on game close.  This avoids some mess of launching overlays while the game is closed.
        
        TODO: I got the game stuck such that launching it from recent apps pops up only the sign in page.
              Click the welcome back toast after closing the game.  Not fixable by force stopping.
              There should be a way to fix this in the overlay's intent.
              Barring that, we could exit the activity if the game is closed.
              Either way, catch the error on activity launch and exit to prevent recurrence on later attempts.
        */
        
        /**
         * Intitialize is called once on game startup.
         * It's goal is to determine what state the GC SDK is in and display the appropriate welcome back toast and/or upsell.
         * The rest of the FTUE workflow continues if the user taps the toast.
         */
        this['onInitialize'] = function() {
            /* TODO: Should we check COR and TOS periodically? */
            /* TODO: Should welcome back toast be shown on first session start instead of init to support games that use WS in background? */
            var dfd = $.Deferred();
            var playerIdDfd = $.Deferred();
            
            reloadInternalState().always(function() {
                /* Update global state if they're logged in.  Reads guestMode, so this needs to be after reload. */
                updateGlobalPlayerID().always(function() {
                    playerIdDfd.resolve();
                });

                var autoConnectDfd = $.Deferred();
                if (autoConnectEnabled) {
                    doAutoConnectChecks().always(function(checksPassed) {
                        if (checksPassed) {
                            gcSettings.setFTUEConnected(true).always(function() {
                                reloadInternalState().always(function() {
                                    autoConnectDfd.resolve();
                                });
                            });
                        } else {
                            autoConnectDfd.resolve();
                        }
                    });
                } else {
                    autoConnectDfd.resolve();
                }
                
                autoConnectDfd.always(function() {
                    var wantToDoUpsell = false; /* TODO: A big pile of logic for checking if we want to upsell now or later. */
                    if (upsellEnabled && wantToDoUpsell) {
                        /* TODO: Show upsell */
                    } else {
                        playerIdDfd.always(function() {
                            gcSettings.getGuestMode().always(function(globalGuestMode) {
                                if (globalGuestMode) {
                                    showWelcomeGuestToast();
                                } else {
                                    showWelcomeBackToast();
                                }
                            });
                        });
                    }
                    dfd.resolve();
                });
            });

            $.when(dfd, playerIdDfd).always(function() {
                initializationDeferred.resolve();
            });
            
            return dfd.promise();
        };
        
        /**
         * Run through the auto-connect checks and resolve(true) if they pass.
         */
        var doAutoConnectChecks = function() {
            var dfd = $.Deferred();
            getNetworkInfo().always(function(networkInfo) {
                var isOnline = networkInfo != null && networkInfo[constants.NativeCallKeys.CONNECTED];
                if (!isOnline) {
                    console.log("FTUE auto-connect failed: not online");
                    dfd.resolve(false);
                }
                getLoggedInStatus().always(function(isLoggedIn) {
                    if (!isLoggedIn) {
                        console.log("FTUE auto-connect failed: not logged in");
                        dfd.resolve(false);
                    }
                    $.when(profilesHandler.getTOSAccepted(), getCORFromProfileServiceAndUpdateCache()).always(function(isTOSAccepted, cor) {
                        reloadInternalState().always(function() {
                            if (!isTOSAccepted || cachedCORIsBlank || cachedCORIsNotSupported) {
                                console.log("FTUE auto-connect failed: TOS:" + isTOSAccepted
                                        + ", COR: " + cor
                                        + ", COR blank:" + cachedCORIsBlank
                                        + ", COR not supported:" + cachedCORIsNotSupported);
                                dfd.resolve(false);
                            } else {
                                console.log("FTUE auto-connect checks passed");
                                dfd.resolve(true);
                            }
                        });
                    });
                });
            });
            return dfd.promise();
        };
        
        this['getInitializationPromise'] = function() {
            return initializationDeferred.promise();
        };
        
        /* ========================================== Incoming FTUE Workflow Hooks ========================================== */
        
        /**
         * Called when the welcome back toast is clicked.
         */
        this['showSignInPage'] = function() {
            gcSettings.getGuestMode().always(function(globalGuestMode) {
                if (globalGuestMode) {
                    getLoggedInStatus().always(function(isLoggedIn) {
                        if (isLoggedIn) {
                            showOverlay(constants.OverlayActionCode.SHOW_TERMS_OF_SERVICE);
                        } else {
                            showOverlay(constants.OverlayActionCode.SHOW_LOGIN);
                        }
                    });
                } else {
                    /* They're not supposed to be able to click sign-in while not in guest mode.  Show something else instead. */
                    showOverlay(constants.OverlayActionCode.SHOW_SUMMARY);
                }
            });
        };

        /**
         * Called when the user wants to sign in or connect.
         */
        this['authenticate'] = function() {
            getLoggedInStatus().always(function(isLoggedInBefore) {
                var doLoginPromise;
                if (isLoggedInBefore) {
                    /* Nothing to do. */
                    doLoginPromise = $.Deferred().resolve().promise();
                } else {
                    doLoginPromise = NativeTransport.callNative({nativeCall: constants.NativeCallTypes.DO_LOGIN});
                }
                
                doLoginPromise.always(function() {
                    getLoggedInStatus().always(function(isLoggedIn) {
                        if (isLoggedIn) {
                            /* TODO: Profile call could be done in parallel with IBA sign-in. */
                            getCORFromProfileServiceAndUpdateCache().always(function(cor) {
                                reloadInternalState().always(function() {
                                    if (cachedCORIsBlank) {
                                        showOverlay(constants.OverlayActionCode.SHOW_COR_SELECTION);
                                    } else if (cachedCORIsNotSupported) {
                                        showCORNotSupported();
                                    } else {
                                        /* Store TOS acceptance after checking COR. */
                                        profilesHandler.acceptTOS().always(function() {
                                            gcSettings.setFTUEConnected(true).always(function() {
                                                /* Logging in, set whispersync to enabled.
                                                 * If changing this, make sure something still handles turning it on after a clean install.
                                                 */
                                                gcSettings.setWhispersyncEnabledUserSetting(true).always(function() {
                                                    reloadInternalState().always(function() {
                                                        /* Start it as soon as logged in since it's global.  Only delay welcome back toast below. */
                                                        var playerInfoPromise = updateGlobalPlayerID();
                                                        gcSettings.getGuestMode().always(function(globalGuestMode) {
                                                            if (!globalGuestMode) {
                                                                playerInfoPromise.always(function() {
                                                                    showWelcomeBackToast();
                                                                });
                                                            } else {
                                                                /* TODO: Something bad or unexpected happened.  show login failed toast? */
                                                            }
                                                        });
                                                    });
                                                });
                                            });
                                        });
                                    }
                                });
                            });
                        } else {
                            /* TODO: show login failed toast */
                        }
                    });
                });
            });
        };
        
        this['updateCOR'] = function(cor) {
            sendCORToProfileServiceAndUpdateCache(cor).always(function() {
                reloadInternalState().always(function() {
                    if (cachedCORIsBlank) {
                        console.error("Failed to update COR.");
                        /* TODO: display error */
                    } else if (cachedCORIsNotSupported) {
                        showCORNotSupported();
                    } else {
                        /* Store TOS acceptance after checking COR. */
                        profilesHandler.acceptTOS().always(function() {
                            gcSettings.setFTUEConnected(true).always(function() {
                                reloadInternalState().always(function() {
                                    gcSettings.getGuestMode().always(function(globalGuestMode) {
                                        if (globalGuestMode) {
                                            /* TODO: display error */
                                        } else {
                                            showWelcomeBackToast();
                                        }
                                    });
                                });
                            });
                        });
                    }
                });
            });
        };

        /* TODO: Call from settings page. */
        /** Called when a user configurable setting has changed that affects FTUE. */
        this['onSettingsChanged'] = function() {
            return reloadInternalState();
        };
        
        /* ========================================== Outgoing FTUE Workflow Calls ========================================== */

        var showWelcomeGuestToast = function() {
            var toastRequest = {};
            toastRequest['type'] = 'WelcomeBack';
            toastRequest['dedupeTypes'] = ['WelcomeBack'];
            toastRequest['target'] = 'WelcomeGuest';
            toastRequest['method'] = 'Not used?';
            toastRequest[constants.ServiceActionCode.ACTION_CODE_KEY] = constants.ServiceActionCode.SHOW_SIGN_IN_PAGE;
            hostinterface.showToast(JSON.stringify(toastRequest));
        };
        
        var showWelcomeBackToast = function() {
            profilesHandler.getProfile(true /* cache only */).always(function(profile) {
                var toastRequest = {};
                toastRequest['type'] = 'WelcomeBack';
                toastRequest['dedupeTypes'] = ['WelcomeBack'];
                toastRequest['target'] = 'WelcomeBack';
                toastRequest['method'] = 'Not used?';
                toastRequest['alias'] = profile.alias;
                /* TODO: should be summary */
                toastRequest[constants.ServiceActionCode.ACTION_CODE_KEY] = constants.ServiceActionCode.SHOW_OVERLAY_PROFILE;
                hostinterface.showToast(JSON.stringify(toastRequest));
            });
        };
        
        var showCORNotSupported = function() {
            showOverlay(constants.OverlayActionCode.SHOW_COR_UNSUPPORTED);
        };
        
        var showOverlay = function(overlayActionCode) {
            var overlayRequest = {};
            overlayRequest[constants.OverlayBindingKeys.OVERLAY_ACTION_CODE_KEY] = overlayActionCode;
            hostinterface.showOverlay(JSON.stringify(overlayRequest));
        };

        /* ========================================== Misc ========================================== */
        
        var getCORFromProfileServiceAndUpdateCache = function() {
            var dfd = $.Deferred();
            profilesHandler.getCORSupported().always(function(corSupported) {
                var corIsBlank = corSupported == null || corSupported == '';
                if (corIsBlank) {
                    /* Service doesn't know.  Leaving the cache with whatever it already has. */
                    dfd.resolve(null);
                } else {
                    gcSettings.setUserCOR(corSupported).always(function() {
                        dfd.resolve(corSupported);
                    });
                }
            });
            return dfd.promise();
        };
        
        var sendCORToProfileServiceAndUpdateCache = function(cor) {
            var dfd = $.Deferred();
            profilesHandler.setCOR(cor).always(function() {
                gcSettings.setUserCOR(cor).always(function() {
                    dfd.resolve();
                });
            });
            return dfd.promise();
        };

        var updateGlobalPlayerID = function() {
            var dfd = $.Deferred();
            /* TODO: add cache.  Also, make sure cache gets cleared if user becomes offline. */
            gcSettings.getGuestMode().always(function(globalGuestMode) {
                if (!globalGuestMode) {
                    getNetworkInfo().always(function(networkInfo) {
                        var isOnline = networkInfo != null && networkInfo[constants.NativeCallKeys.CONNECTED];
                        if (isOnline) {
                            profilesHandler.getProfile().always(function(profile) {
                                console.log("Profiles response: " + JSON.stringify(profile));
                                if (profile != null) {
                                    gcSettings.setPlayerId(profile.playerId).always(function() {
                                        dfd.resolve();
                                    });
                                } else {
                                    dfd.resolve();
                                    /* TODO: Leave it as-is? */
                                }
                            });
                        } else {
                            /* TODO: Leave it as-is? */
                            dfd.resolve();
                        }
                    });
                } else {
                    gcSettings.setPlayerId(null).always(function() {
                        profilesHandler.clearCachedProfile().always(function() {
                            dfd.resolve();
                        });
                    });
                }
            });
            return dfd.promise();
        };

        /* ========================================== Native calls to get stuff ========================================== */
        
        var getNetworkInfo = function() {
            return NativeTransport.callNative({nativeCall: constants.NativeCallTypes.GET_NETWORK_INFO});
        };
        
        var getLoggedInStatus = function() {
            var loggedInPromise = NativeTransport.callNative({nativeCall: constants.NativeCallTypes.GET_LOGGED_IN_STATUS});
            var dfd = $.Deferred();
            loggedInPromise.always(function(isLoggedInResult) {
                var loggedIn = false;
                if (isLoggedInResult != null) {
                    isLoggedIn = isLoggedInResult[constants.NativeCallKeys.LOGGED_IN_STATUS];
                }
                console.log("FTUE: isLoggedIn=" + isLoggedIn);
                dfd.resolve(isLoggedIn);
            });
            return dfd.promise();
        };

        /* ========================================== Handler stuff ========================================== */
        
        this['getHandledType'] = function() {
            return requestTypeHandled;
        };
        
        this['handleRequest'] = function(request) {
            var dfd = $.Deferred();
            
            if (!(request instanceof Request)) {
                console.log('FTUEHandler: handleRequest: input was not of type Request');
                /* TODO: Revise FAILURE result code */
                var result = new Result(constants.NativeCallResultCode.REQUEST_ERROR, {});
                dfd.resolve(result);
            } else {
                console.log('FTUEHandler: handleRequest: message is: ' + JSON.stringify(request.getType()));
                var requestType = request.getType();
                var requestParams = request.getParams();
                var actionCode = requestParams['action'];
                
                var reply = {};
                switch (actionCode) {
                    case 'logout':
                        var logoutPromise = NativeTransport.callNative({nativeCall: constants.NativeCallTypes.DO_LOGOUT});
                        logoutPromise.always(function() {
                            reloadInternalState().always(function() {
                                dfd.resolve();
                            });
                        });
                        break;
                    case 'login':
                        thisHandler.showSignInPage();
                        break;
                    default:
                        console.log('FTUEHandler: handleRequest: actionCode not supported: ' + actionCode);
                        var result = new Result(constants.NativeCallResultCode.REQUEST_ERROR, {});
                        dfd.resolve(result);
                }
            }
            
            console.log('AchievementHandler: handleRequest: end of call');
            return dfd.promise();
        };
    };
    
    return constructor;
}());
console.log("FTUEHandler loaded.");
