/*
 * Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 * AMAZON.COM CONFIDENTIAL
 */

ProfilesHandler = (function() {
    var constructor = function(profileService, profileCache) {
        var requestTypeHandled = 'ProfileRequest';
        var thisHandler = this;
        
        this['getHandledType'] = function() {
            return requestTypeHandled;
        };
        
        /**
         * Returns a promise for getting COR by attempting a network call.
         * Callback returns whether or not the COR is supported, null if unknown.
         */
        this['getCORSupported'] = function() {
            var dfd = $.Deferred();
            dfd.resolve(true); /* TODO */
            return dfd.promise();
        };
        
        /**
         * Returns a promise for updating COR by attempting a network call.
         * Callback returns whether or not the COR is supported.
         */
        this['setCOR'] = function(cor) {
            var dfd = $.Deferred();
            dfd.resolve(cor == 'valid'); /* TODO */
            return dfd.promise();
        };
        
        /**
         * Returns a promise for attempting to refresh the cached TOS Acceptance by attempting a network call.
         * Callback returns whether or not the TOS is accepted, null if unknown.
         */
        this['getTOSAccepted'] = function() {
            var dfd = $.Deferred();
            dfd.resolve(false); /* TODO */
            return dfd.promise();
        };
        
        /**
         * Returns a promise for acccepting TOS by attempting a network call.
         * Callback returns nothing.
         */
        this['acceptTOS'] = function() {
            var dfd = $.Deferred();
            dfd.resolve(); /* TODO */
            return dfd.promise();
        };

        this['clearCachedProfile'] = function() {
            return profileCache.cacheProfile('SELF', {});
        };
        
        /* TODO: Remove */
        this['getProfile'] = function(isCacheOnly) {
            var dfd = $.Deferred();
            if (isCacheOnly) {
                profileCache.getCachedProfile('SELF').always(function(result, resultCode) {
                    console.log('ProfileHandler: getProfile result: ' + JSON.stringify(result));
                    dfd.resolve(result);
                });
            } else {
                profileService.getProfile().always(function(result, resultCode) {
                    console.log('ProfileHandler: getProfile result: ' + JSON.stringify(result));

                    if (resultCode != constants.NativeCallResultCode.SUCCESS) {
                        console.log('ProfileHandler: getProfile result: SERVICE RESULT NOT OK');
                        profileCache.getCachedProfile('SELF').always(function(result, resultCode) {
                            console.log('ProfileHandler: getProfile promise: ' + JSON.stringify(result));
                            dfd.resolve(result);
                        });
                    } else {
                        var cachePromise = profileCache.cacheProfile('SELF', result);
                        cachePromise.always(function() {
                            console.log('UPDATED PROFILE CACHE');
                            /* Delay resolving until after cache update: Welcome back toast logic will immediately try to read the cache. */
                            dfd.resolve(result);
                        });
                    }
                });
            }
            return dfd.promise();
        };

        this['handleRequest'] = function(request) {
            /*
             * Step 1: Ensure request is of the appropriate type
             * Step 2: Depending on the request's action code, call profile service proxy
             * Step 3: Perform any caching or cache lookup as necessary
             * Step 4: Return data
             */
            var dfd = $.Deferred();
            
            if (!(request instanceof Request)) {
                console.log('ProfileHandler: handleRequest: input was not of type Request');
                var result = new Result(constants.NativeCallResultCode.REQUEST_ERROR, {});
                dfd.resolve(result);
            } else {
                console.log('ProfileHandler: handleRequest: message is: ' + JSON.stringify(request.getType()));
                var requestType = request.getType();
                var requestParams = request.getParams();
                var actionCode = requestParams['action'];
                
                var reply = {};
                switch (actionCode) {
                    case 'getProfile':
                        thisHandler.getProfile(requestParams.cacheOnly).always(function(profile, resultCode) {
                            dfd.resolve(new Result(resultCode, profile));
                        });
                        break;
                    case 'updateProfile':
                        profileService.updateProfile(requestParams.profile).always(function(profile, resultCode) {
                            dfd.resolve(new Result(resultCode, profile));
                        });
                        break;
                    case 'updateProfileHidden':
                        profileService.updateProfileHidden(requestParams.hidden).always(function(result, resultCode) {
                            dfd.resolve(new Result(resultCode, result));
                        });
                        break;
                    case 'getAvatars':
                        /* TODO: support caching and/or retries?  If we add caching, edit profile would need to work with that too. */
                        if (requestParams.cacheOnly) {
                            dfd.resolve(new Result(constants.NativeCallResultCode.SUCCESS, {}));
                        } else {
                            profileService.getAvatars().always(function(avatars, resultCode) {
                                dfd.resolve(new Result(resultCode, avatars));
                            });
                        }
                        break;
                    default:
                        console.log('ProfileHandler: handleRequest: actionCode not supported: ' + actionCode);
                        var result = new Result(constants.NativeCallResultCode.REQUEST_ERROR, {});
                        dfd.resolve(result);
                }
            }
            
            console.log('ProfileHandler: handleRequest: end of call');
            return dfd.promise();
        };
    };

    return constructor;
}());
console.log("ProfilesHandler loaded.");
