/*
 * Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 * AMAZON.COM CONFIDENTIAL
 */

GCSettings = (function() {
    var constructor = function() {
        var GLOBAL_GUEST_MODE = constants.NativeCallKeys.GUEST_MODE;
        var GLOBAL_PLAYER_ID = constants.NativeCallKeys.PLAYER_ID;
        var SETTINGS_FTUE_CONNECTED = "ftue:connected";
        var SETTINGS_USER_WHISPERSYNC_ENABLED = "user:whispersyncEnabled";
        var SETTINGS_USER_COR = "user:cor";
        
        /* ========================== Global State ========================== */
        /* These methods are backed by a global Map<String, String> in java memory and are not persisted long term. */

        /**
         * This is the boolean to check before making write service calls that send data from the device to the cloud.
         * Don't attempt write service calls when in guest mode. (Read-only is ok)
         * When guestMode==true they may not have accepted the TOS, or they may not have a valid COR.
         * Note that the IBA SDK may be able to return valid tokens even if guestMode == true.
         */
        this['getGuestMode'] = function() {
            var dfd = $.Deferred();
            getGlobalStateAsAString(GLOBAL_GUEST_MODE, "true" /* default */).always(function(resultString) {
                dfd.resolve(resultString == "true");
            });
            return dfd.promise();
        };
        
        /* This should only be called by FTUE, and is derived from other settings and state. */
        this['setGuestMode'] = function(guestMode) {
            return setGlobalState(GLOBAL_GUEST_MODE, guestMode);
        };
        
        /**
         * Player ID, if available.  May not be immediately available after initialize.
         */
        this['getPlayerId'] = function() {
            return getGlobalStateAsAString(GLOBAL_PLAYER_ID, null /* default */);
        };
        
        /* This should only be called by FTUE. */
        this['setPlayerId'] = function(playerId) {
            return setGlobalState(GLOBAL_PLAYER_ID, playerId);
        };

        /* ========================== FTUE ========================== */
        
        /* FTUE setting for whether or not the FTUE has been completed and the game is connected. */
        this['getFTUEConnected'] = function() {
            return getSingleSetting(SETTINGS_FTUE_CONNECTED);
        };
        
        this['setFTUEConnected'] = function(updatedFTUEConnected) {
            return setSingleSetting(SETTINGS_FTUE_CONNECTED, updatedFTUEConnected);
        };

        /* User's country of residence. */
        this['getUserCOR'] = function() {
            /* TODO: profiles network call.  migrate this over there? */
            return getSingleSetting(SETTINGS_USER_COR);
        };
        
        this['setUserCOR'] = function(udpatedCOR) {
            /* TODO: profiles network call.  migrate this over there? */
            return setSingleSetting(SETTINGS_USER_COR, udpatedCOR);
        };
        
        /* ========================== Settings page ========================== */
        
        this['getWhispersyncEnabledUserSetting'] = function() {
            return getSingleSetting(SETTINGS_USER_WHISPERSYNC_ENABLED);
        };
        
        this['setWhispersyncEnabledUserSetting'] = function(enabled) {
            return setSingleSetting(SETTINGS_USER_WHISPERSYNC_ENABLED, enabled);
        };
        
        /* ========================== Util ========================== */

        var getSingleSetting = function(settingName) {
            var request = {};
            request[constants.NativeCallKeys.PRIMARY_KEY] = settingName;
            var promise = NativeTransport.callNative({nativeCall: constants.NativeCallTypes.GET_SETTING, args: request});
            
            /* Wrap response in a deferred to unpack structure: {value: 'foo'} */
            var dfd = $.Deferred();
            promise.always(function(result) {
                if (result == null) {
                    dfd.resolve(null);
                } else {
                    dfd.resolve(result.value);
                }
            });

            return dfd.promise();
        };
        
        var setSingleSetting = function(settingName, settingValue) {
            var request = {};
            request[constants.NativeCallKeys.PRIMARY_KEY] = settingName;
            request[constants.NativeCallKeys.SECONDARY_KEY] = settingName;
            request[constants.NativeCallKeys.VALUE] = {value: settingValue};
            return NativeTransport.callNative({nativeCall: constants.NativeCallTypes.PUT_SETTING, args: request});
        };
        
        /* Warning: all values returned are strings. */
        var getGlobalStateAsAString = function(settingName, defaultIfNotSet) {
            var request = {};
            request[constants.NativeCallKeys.PRIMARY_KEY] = settingName;
            var promise = NativeTransport.callNative({nativeCall: constants.NativeCallTypes.GET_GLOBAL_STATE, args: request});
            
            /* Wrap response in a deferred to unpack response structure: {value: 'foo'} */
            var dfd = $.Deferred();
            promise.always(function(result) {
                if (result == null || result.value == null) {
                    dfd.resolve(defaultIfNotSet);
                } else {
                    dfd.resolve(result.value);
                }
            });

            return dfd.promise();
        };

        var setGlobalState = function(settingName, settingValue) {
            var request = {};
            request[constants.NativeCallKeys.PRIMARY_KEY] = settingName;
            request[constants.NativeCallKeys.VALUE] = settingValue;
            return NativeTransport.callNative({nativeCall: constants.NativeCallTypes.PUT_GLOBAL_STATE, args: request});
        };
    };
    
    return constructor;
}());
console.log("GCSettings loaded.");
