/*
 * Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 * AMAZON.COM CONFIDENTIAL
 */

TimePlayedHandler = (function() {
    var constructor = function(initializationPromise) {
        var requestTypeHandled = 'TimePlayedRequest';
        var thisHandler = this;

        /* Settings key name for time played data. */
        var SETTINGS_TIME_PLAYED = 'timePlayed';
        
        /* Key name for timestamp for last start.  The timestamp is null after a stop. */
        var LAST_START_KEY = 'sessionStart';
        
        /**
         * Key name for total time played in millis.
         * The associated value is only changed when reading/writing to the cloud,
         * and approximates the total time played stored in the cloud.
         */
        var CLOUD_TIME_PLAYED_KEY = 'cloudTimePlayed';
        
        /**
         * Key name for total time played in millis that hasn't been uploaded.
         * The associated value accumulates during session end, and later gets transferred to pending, then cloud.
         */
        var LOCAL_TIME_PLAYED_KEY = 'localTimePlayed';
        
        /* Key name for timestamp for last time played. Updated on start and stop. */
        var LAST_TIME_PLAYED_KEY = 'lastTimePlayed';

        /**
         * Key name for total time played in millis that may or may not have been uploaded.
         */
        var LOCAL_TIME_PLAYED_PENDING_UPLOAD_KEY = 'localTimePlayedPendingUpload';

        /* Key name for timestamp for last time played, that may or may not have been uploaded. */
        var LAST_TIME_PLAYED_PENDING_UPLOAD_KEY = 'lastTimePlayedPendingUpload';
        
        /* To reduce server load, don't upload until the local time reaches this amount. */
        var MINIMUM_UPLOAD_AMOUNT_MILLIS = 20000;

        /* TODO: Do we want to complain if not started/closed between inits, to remind developers that they should add the hooks?
                 It's easy to forget to add them.
                 This can be done by storing the last time init/start/stop happened (concurrency issues),
                 and complaining on init if the others aren't getting called. */

        /* TODO: Handle concurrency during read/write. */
        
        /* TODO: Reduce server load: Don't upload delta if not enough time has been accumulated, for example 20 seconds. */
        /* TODO: Discard very short sessions.  The device sometimes start/stops a session for well under a second several times in a row, while sleeping. */
        /* TODO: Keep time tracking during overlays or not? */

        /* ==================================== Event Hooks ==================================== */

        /* There's no guarantee that start and stop will be called in a sane order. */
        this['startSession'] = function() {
            console.log("TimePlayed: Start");
            var startTime = new Date().getTime();
            getLocalTimePlayedData().always(function(data) {
                if (data[LAST_START_KEY] != null) {
                    /* Note for log scrubbing: this should be a warning to the developer. */
                    console.warn("onResume() called twice in a row: please call onPause()");
                }
                data[LAST_START_KEY] = startTime;
                
                updateLastTimePlayed(data, startTime);
                setLocalTimePlayedData(data).always(function() {
                    initializationPromise.always(function() {
                        syncTimePlayed();
                    });
                });
            });
        };

        this['stopSession'] = function() {
            console.log("TimePlayed: Stop");
            getLocalTimePlayedData().always(function(data) {
                /**
                 * Get the endTime after fetching local data so it should be larger.
                 * isLastStartValid will reject the start time if this isn't true.
                 */
                var endTime = new Date().getTime();
                
                if (isLastStartValid(data[LAST_START_KEY], endTime)) {
                    var delta = endTime - data[LAST_START_KEY];
                    if (data[LOCAL_TIME_PLAYED_KEY] == null) {
                        data[LOCAL_TIME_PLAYED_KEY] = 0;
                    }
                    data[LOCAL_TIME_PLAYED_KEY] += delta;
                }
                data[LAST_START_KEY] = null;
                
                updateLastTimePlayed(data, endTime);
                setLocalTimePlayedData(data);
                /* No need to sync from cloud here, it will be done on next start. */
            });
        };
        
        /* ==================================== Getters ==================================== */
        
        this['getTimePlayedData'] = function() {
            var dfd = $.Deferred();
            getLocalTimePlayedData().always(function(data) {
                var result = {};
                if (data != null) {
                    result.lastTimePlayed = data[LAST_TIME_PLAYED_KEY];
                    
                    var localTimePlayed = data[LOCAL_TIME_PLAYED_KEY] || 0;
                    var localTimePlayedPendingUpload = data[LOCAL_TIME_PLAYED_PENDING_UPLOAD_KEY] || 0;
                    var cloudTimePlayed = data[CLOUD_TIME_PLAYED_KEY] || 0;
                    result.totalTimePlayed = localTimePlayed + localTimePlayedPendingUpload + cloudTimePlayed;
                }
                dfd.resolve(result);
            });
            return dfd.promise();
        };
        
        /* ==================================== Sync From Cloud ==================================== */
        
        var fetchTimePlayedFromCloud = function() {
            /* TODO: To reduce server load, don't check if we've already recently checked. */
            var dfd = $.Deferred();
            getPackageName().done(function(packageName) {
                var args = {packageName: packageName};
                var getPromise = NativeTransport.serviceCall({target: "TimePlayedService", method: "getGameTimePlayed", args: args});
                getPromise.always(function(remoteData) {
                    if (remoteData == null) {
                        dfd.resolve();
                        return;
                    }
                    
                    getLocalTimePlayedData().always(function(data) {
                        var delta = remoteData.totalTimePlayed - data[CLOUD_TIME_PLAYED_KEY];
                        console.log("TimePlayed: updated from cloud, delta=" + delta);
                        data[CLOUD_TIME_PLAYED_KEY] = remoteData.totalTimePlayed;
                        updateLastTimePlayed(data, remoteData.lastTimePlayed);
                        setLocalTimePlayedData(data).always(function() {
                            dfd.resolve();
                        });
                    });
                });
            }).fail(function() {
                console.error("Unable to fetch time played from cloud: no package name");
                dfd.reject();
            });
            return dfd.promise();
        };
        
        /* ==================================== Sync To Cloud ==================================== */
        
        var uploadTimePlayedToCloud = function() {
            /* TODO: serialize all calls as required for the idempotence logic. */
            var dfd = $.Deferred();
            moveLocalTimeToPendingIfNeeded().always(function() {
                uploadPendingDataIfNeeded().always(function() {
                    dfd.resolve();
                });
            });
            return dfd.promise();
        };
        
        /** Returns true if data was modified. */
        var moveLocalTimeToPendingIfNeeded = function(data) {
            var dfd = $.Deferred();
            getLocalTimePlayedData().always(function(data) {
                if (data == null) {
                    dfd.resolve();
                } else if (hasPendingTimePlayed(data)) {
                    /* There's already pending data. */
                    dfd.resolve();
                } else if (data[LOCAL_TIME_PLAYED_KEY] == null || data[LOCAL_TIME_PLAYED_KEY] < MINIMUM_UPLOAD_AMOUNT_MILLIS) {
                    /* There's nothing worth uploading yet. */
                    dfd.resolve();
                } else if (data[LAST_TIME_PLAYED_KEY] == null || data[LAST_TIME_PLAYED_KEY] <= 0) {
                    /* Missing last time played. */
                    dfd.resolve();
                } else {
                    data[LOCAL_TIME_PLAYED_PENDING_UPLOAD_KEY] = data[LOCAL_TIME_PLAYED_KEY];
                    data[LOCAL_TIME_PLAYED_KEY] = 0;
                
                    /* Ensure next upload has a different last time played than the previous.
                       If it's the same, then the server will assume the request is a retry and discard the data.
                       This can be caused by setting the device time to the future and back,
                       which will make last time played stop advancing for a while. */
                    if (data[LAST_TIME_PLAYED_KEY] == data[LAST_TIME_PLAYED_PENDING_UPLOAD_KEY]) {
                        data[LAST_TIME_PLAYED_PENDING_UPLOAD_KEY] = data[LAST_TIME_PLAYED_KEY] + 1;
                    } else {
                        data[LAST_TIME_PLAYED_PENDING_UPLOAD_KEY] = data[LAST_TIME_PLAYED_KEY];
                    }
                    
                    setLocalTimePlayedData(data).always(function() {
                        dfd.resolve();
                    });
                }
            });
            return dfd.promise();
        };
        
        var uploadPendingDataIfNeeded = function() {
            var dfd = $.Deferred();
            /* TODO: send pending data, and purge when done. */
            getLocalTimePlayedData().always(function(data) {
                if (!hasPendingTimePlayed(data)) {
                    dfd.resolve();
                    return;
                }
                
                getPackageName().done(function(packageName) {
                    var args = {
                        packageName: packageName,
                        totalTimePlayed: data[LOCAL_TIME_PLAYED_PENDING_UPLOAD_KEY],
                        lastTimePlayed: data[LAST_TIME_PLAYED_PENDING_UPLOAD_KEY]
                    };
                    var updatePromise = NativeTransport.serviceCall({target: "TimePlayedService", method: "updateGameTimePlayed", args: args});
                    updatePromise.done(function() {
                        movePendingTimeToCloud(data).always(function() {
                            dfd.resolve();
                        });
                    }).fail(function() {
                        /* TODO: There's edge cases here where the server rejects invalid requests and we keep trying to send.
                                 One possible catch-all is to stop re-sending after a certain number of attempts and enough time has passed. */
                        console.warn("Failed to upload time played data");
                        dfd.resolve();
                    });
                }).fail(function() {
                    console.error("Unable to upload time played to cloud: no package name");
                    dfd.reject();
                });
            });
            return dfd.promise();
        };
        
        /* TODO: Might need to cache time played until cloud catches up. */
        var movePendingTimeToCloud = function(dataSent) {
            var dfd = $.Deferred();
            getLocalTimePlayedData().always(function(data) {
                if (data[LAST_TIME_PLAYED_PENDING_UPLOAD_KEY] != dataSent[LAST_TIME_PLAYED_PENDING_UPLOAD_KEY]) {
                    console.error("Last time played uploaded no longer matches what is cached locally!");
                    dfd.resolve();
                    return;
                }
                
                console.log("TimePlayed: uploaded " + data[LOCAL_TIME_PLAYED_PENDING_UPLOAD_KEY] + " to cloud.");
                data[CLOUD_TIME_PLAYED_KEY] += data[LOCAL_TIME_PLAYED_PENDING_UPLOAD_KEY];
                data[LOCAL_TIME_PLAYED_PENDING_UPLOAD_KEY] = 0;
                /* Leaving the pending last time played as-is, since it is needed for checks during the next upload. */
                
                setLocalTimePlayedData(data).always(function() {
                    /* No error handling needed, if this fails we'll retry and it will get deduped on the server. */
                    dfd.resolve();
                });
            });
            return dfd.promise();
        };
        
        var hasPendingTimePlayed = function(data) {
            var pendingTime = data[LOCAL_TIME_PLAYED_PENDING_UPLOAD_KEY];
            return pendingTime != null && pendingTime > 0;
        };
        
        /* ==================================== Misc ==================================== */

        /**
         * TODO: Document how the idempotence mechanism works.
         * See GameTimePlayedManager.java in GameCircleGameDataService.
         */
        var syncTimePlayed = function() {
            /* TODO: check online */
            /* Serializing to reduce probability of concurrency issues until that's fixed. */
            /* TODO: switch to a real serializer and get rid of this method. */
            uploadTimePlayedToCloud().always(function() {
                fetchTimePlayedFromCloud();
            });
        };

        var updateLastTimePlayed = function(data, newTimePlayed) {
            if (data[LAST_TIME_PLAYED_KEY] == null || data[LAST_TIME_PLAYED_KEY] < newTimePlayed) {
                data[LAST_TIME_PLAYED_KEY] = newTimePlayed;
            }
        };
        
        var isLastStartValid = function(lastStart, now) {
            if (lastStart == null) {
                /* Note for log scrubbing: this should be a warning to the developer. */
                console.warn("onPause() called without first calling onResume()");
                return false;
            }
            
            if (lastStart <= 0) {
                console.warn("Bad start time " + lastStart + ", discarding this session's time tracking.");
                return false;
            }
            
            if (lastStart > now) {
                console.warn("Start time is after end time, discarding this session's time tracking.");
                return false;
            }
            
            return true;
        };
        
        /* ==================================== Native calls ==================================== */
        
        /**
         * Returns a promise for fetching whether or not FTUE has been completed, from cache.
         * Callback parameter is a boolean.
         */
        var getLocalTimePlayedData = function() {
            var request = {};
            request[constants.NativeCallKeys.PRIMARY_KEY] = SETTINGS_TIME_PLAYED;
            return NativeTransport.callNative({nativeCall: constants.NativeCallTypes.GET_SETTING, args: request});
        };

        /**
         * Returns a promise for updating the cache.
         */
        var setLocalTimePlayedData = function(data) {
            var request = {};
            request[constants.NativeCallKeys.PRIMARY_KEY] = SETTINGS_TIME_PLAYED;
            request[constants.NativeCallKeys.SECONDARY_KEY] = SETTINGS_TIME_PLAYED;
            request[constants.NativeCallKeys.VALUE] = data;
            return NativeTransport.callNative({nativeCall: constants.NativeCallTypes.PUT_SETTING, args: request});
        };
        
        var getPackageName = function() {
            var dfd = $.Deferred();
            NativeTransport.callNative({nativeCall: constants.NativeCallTypes.GET_PACKAGE_NAME}).always(function(result) {
                if (result != null) {
                    var packageName = result[constants.BindingKeys.PACKAGE_NAME_KEY];
                    if (packageName != null && packageName != '') {
                        dfd.resolve(packageName);
                        return;
                    }
                }
                dfd.reject();
            });
            return dfd.promise();
        };
    
        /* ==================================== Handler ==================================== */

        this['getHandledType'] = function() {
            return requestTypeHandled;
        };
        
        this['handleRequest'] = function(request) {
            /*
             * Step 1: Ensure request is of the appropriate type
             * Step 2: Depending on the request's action code, call profile service proxy
             * Step 3: Perform any caching or cache lookup as necessary
             * Step 4: Return data
             */
            var dfd = $.Deferred();
            
            if (!(request instanceof Request)) {
                console.log('TimePlayedHandler: handleRequest: input was not of type Request');
                var result = new Result('FAILURE', {});
                dfd.resolve(result);
            } else {
                console.log('TimePlayedHandler: handleRequest: message is: ' + JSON.stringify(request.getType()));
                var requestType = request.getType();
                var requestParams = request.getParams();
                var actionCode = requestParams['action'];
                
                var reply = {};
                switch (actionCode) {
                    case 'getTimePlayed':
                        /* TODO: add support for cacheOnly to getTimePlayedData.  It could sync from cloud for the non-cache call. */
                        thisHandler.getTimePlayedData(requestParams.cacheOnly).always(function(timePlayedData) {
                            dfd.resolve(new Result(constants.NativeCallResultCode.SUCCESS, timePlayedData));
                        });
                        break;
                    default:
                        console.log('TimePlayedHandler: handleRequest: actionCode not supported: ' + actionCode);
                        var result = new Result('FAILURE', {});
                        dfd.resolve(result);
                }
            }
            
            console.log('TimePlayedHandler: handleRequest: end of call');
            return dfd.promise();
        };
    };
    
    return constructor;
}());
console.log("TimePlayedHandler loaded.");
